static char rcsid[] = "@(#)$Id: aliases.c,v 1.9 2014/11/22 10:03:21 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.9 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> (was hurtta+elm@posti.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/


#include "def_alias.h"
#include "s_me.h"

DEBUG_VAR(Debug,__FILE__,"alias");

#define ALIASES_map_magic	0xF602
#define ALIASES_node_magic      0xF606

static struct aliases_map_node {
    unsigned short            magic;    /* ALIASES_node_magic */

    struct string         *alias_name;
    struct address_alias  *alias;

    int                   idx;

    struct aliases_map_node  * smaller; /* LEFT */
    struct aliases_map_node  * bigger;  /* RIGHT */

    struct aliases_map_node  * bad;     /* NO COMPARE */
} * give_aliases_map_node P_((struct aliases_map_node **root, 
			      const struct string * alias_name,
			      int create_falg));

static struct aliases_map_node * give_aliases_map_node(root,alias_name,create_flag)
     struct aliases_map_node **root;
     const struct string * alias_name;
     int create_flag;
{
    struct aliases_map_node *ptr = *root;
    int r;

    if (!ptr) {

	if (!create_flag) {

	    DPRINT(Debug,9,(&Debug,
			    "give_aliases_map_node: \"%S\" not found\n",
			    alias_name));

	    return NULL;
	}

	ptr = safe_malloc(sizeof(*ptr));

        bzero((void *)ptr, sizeof(*ptr));

        ptr->magic = ALIASES_node_magic;

	ptr->alias_name = dup_string(alias_name);
	ptr->idx        = -1;
	ptr->smaller    = NULL;
        ptr->bigger     = NULL;
        ptr->bad        = NULL;
	
	ptr->alias      = NULL;

	
	*root = ptr;

	DPRINT(Debug,14,(&Debug,
			 "give_aliases_map_node: Added \"%S\" as %p\n",
			 alias_name,ptr));


	return ptr;
    }

    if (ptr->magic != ALIASES_node_magic)
	panic("ALIASES PANIC",__FILE__,__LINE__,"give_aliases_map_node",
              "Bad magic number",0);

    r = string_cmp(alias_name,ptr->alias_name,
		   999 /* == Values not comparable */);

    DPRINT(Debug,15,(&Debug,
		     "give_aliases_map_node:     Travelling \"%S\" as %p   ",
		     ptr->alias_name,ptr));

    switch(r) {

    case 0:

	DPRINT(Debug,15,(&Debug," (%d)   FOUND \"%S\"\n",
			 r,alias_name));

        return ptr;
	
    case -1:

	DPRINT(Debug,15,(&Debug," (%d)   > \"%S\"\n",
			 r,alias_name));

	return give_aliases_map_node(& (ptr->bigger), alias_name, create_flag);

    case 1:

	DPRINT(Debug,15,(&Debug," (%d)   < \"%S\"\n",
			 r,alias_name));


	return give_aliases_map_node(& (ptr->smaller), alias_name, create_flag);

    case 999:

	DPRINT(Debug,15,(&Debug," (%d)   <> \"%S\"\n",
			 r,alias_name));


        return give_aliases_map_node(& (ptr->bad), alias_name, create_flag);
    }
    
    panic("MBX VIEW PANIC",__FILE__,__LINE__,"give_aliases_map_node",
	  "Unexpected return from string_cmp()",0);

    return NULL;    
}

const struct aliases_map_node * lookup_aliases_map_node P_((const struct aliases_map_node *root, 
							    const struct string * alias_name));

const struct aliases_map_node * lookup_aliases_map_node(root,alias_name)
     const struct aliases_map_node *root;
     const struct string * alias_name;
{
    int r;

    if (!root) {

	DPRINT(Debug,9,(&Debug,
			"lookup_aliases_map_node: \"%S\" not found\n",
			alias_name));
	
	return NULL;
    }

    if (root->magic != ALIASES_node_magic)
	panic("ALIASES PANIC",__FILE__,__LINE__,"lookup_aliases_map_node",
              "Bad magic number",0);

    r = string_cmp(alias_name,root->alias_name,
		   999 /* == Values not comparable */);


    DPRINT(Debug,15,(&Debug,
		     "lookup_aliases_map_node:     Travelling \"%S\" as %p   ",
		     root->alias_name,root));



    switch(r) {

    case 0:

	DPRINT(Debug,15,(&Debug," (%d)   FOUND \"%S\"\n",
			 r,alias_name));


        return root;
	
    case -1:

	DPRINT(Debug,15,(&Debug," (%d)   > \"%S\"\n",
			 r,alias_name));


	return lookup_aliases_map_node(root->bigger, alias_name);

    case 1:

	DPRINT(Debug,15,(&Debug," (%d)   < \"%S\"\n",
			 r,alias_name));



	return lookup_aliases_map_node(root->smaller, alias_name);

    case 999:

	DPRINT(Debug,15,(&Debug," (%d)   <> \"%S\"\n",
			 r,alias_name));


        return lookup_aliases_map_node(root->bad, alias_name);
    }
    
    panic("MBX VIEW PANIC",__FILE__,__LINE__,"lookup_aliases_map_node",
	  "Unexpected return from string_cmp()",0);

    return NULL;    
}



struct aliases_map {
    unsigned short magic;        /*  ALIASES_map_magic */

    struct aliases_map_node  * root;

    struct aliases_map_node  **vector;
    int                        vector_len;
};

static void free_node_tree P_((struct aliases_map *map,
			       struct aliases_map_node **root));
static void free_node_tree(map,root)
     struct aliases_map *map;
     struct aliases_map_node **root;
{
    if ((*root)->magic != ALIASES_node_magic)
	panic("ALIASES PANIC",__FILE__,__LINE__,"free_node_tree",
              "Bad magic number",0);

    if ((*root)->alias_name)
	free_string(& ((*root)->alias_name));

    if ((*root)->alias)
	free_address_alias(& ((*root)->alias));

    if ((*root)->idx >= 0) {

	if ((*root)->idx >= map->vector_len)
	    panic("ALIASES PANIC",__FILE__,__LINE__,"free_node_tree",
		 "bad index",0); 
			       
	if ((*root) != map->vector[(*root)->idx])
	    panic("ALIASES PANIC",__FILE__,__LINE__,"free_node_tree",
		  "bad vector",0); 

	map->vector[(*root)->idx] = NULL;
    }
    (*root)->idx = -1;


    if ((*root)->smaller)
	free_node_tree(map, & ((*root)->smaller));
    if ((*root)->bigger)
	free_node_tree(map, & ((*root)->bigger));
    if ((*root)->bad)
	free_node_tree(map, & ((*root)->bad));

    (*root)->magic = 0; /* Invalidate */
    free(*root);
    *root = NULL;
}

void free_aliases_map(map)
     struct aliases_map **map;
{

    if (ALIASES_map_magic != (*map)->magic)
	panic("URL PANIC",__FILE__,__LINE__,"free_aliases_map",
	      "Bad magic number",0);

    if ((*map)->root)
	free_node_tree(*map, & ((*map)->root));

    if ((*map)->vector) {
	free((*map)->vector);
	(*map)->vector = NULL;
    }
    (*map)->vector_len = 0;

    (*map)->magic = 0;      /* Invalidate */
    free(*map);
    *map = NULL;
}


struct aliases_map * new_aliases_map()
{
    struct aliases_map *ret = safe_malloc(sizeof(*ret));

    /* bzero is defined hdrs/elm_defs.h */
    bzero((void *)ret,sizeof (*ret));

    ret->magic         = ALIASES_map_magic;

    ret->root          = NULL;
    ret->vector        = NULL;
    ret->vector_len    = 0;

    return ret;
}

static void fill_vector P_((struct aliases_map_node *root,
			    struct aliases_map_node **vector,
			    int alloced_vector,
			    int *idx));

static void fill_vector(root,vector,alloced_vector,idx)
     struct aliases_map_node *root;
     struct aliases_map_node **vector;
     int alloced_vector;
     int *idx;
{
    if (*idx >= alloced_vector)
	panic("ALIASES PANIC",__FILE__,__LINE__,"fill_vector",
	      "Overflow",0);

    if (root->magic != ALIASES_node_magic)
	panic("ALIASES PANIC",__FILE__,__LINE__,"fill_vector",
              "Bad magic number",0);

    vector[*idx] = root;
    root->idx    = *idx;
    (*idx)++;

    if (root->smaller)
	fill_vector(root->smaller,vector,alloced_vector,idx);

    if (root->bigger)
	fill_vector(root->bigger,vector,alloced_vector,idx);

    if (root->bad)
	fill_vector(root->bad,vector,alloced_vector,idx);

}


/*  Format:
 *
 *  [alias]         Alias name
 *  Comment:        Comment text
 *
 *  Firstname:     name
 *  Lastname:      name
 *  Address:       One address
 *
 *  List:          addr-list

 *  Phrase:        phrase text
 *  Group:         alias-list (',' separated)
 *
 */


#define ALIAS_TOKEN   	"[alias]"
#define COMMENT_TOKEN	"Comment"
#define FIRSTNAME_TOKEN	"Firstname"
#define LASTNAME_TOKEN  "Lastname"
#define ADDRESS_TOKEN 	"Address"
#define LIST_TOKEN	"List"
#define PHRASE_TOKEN    "Phrase"
#define GROUP_TOKEN     "Group"

struct aliases_map * load_aliases_map(filename,errors)
     const char *filename;
     int *errors;
{
    struct aliases_map *ret = NULL;

    charset_t  cs  = system_charset;    /* Charset of file */
    int err = can_open(filename,"r");
    FILE * f;
    struct aliases_map_node *map_root = NULL;
    struct aliases_map_node * current_node = NULL;
    int alias_count = 0;

    struct aliases_map_node **vector = NULL;
    int vector_len = 0;

    char * buf = NULL;
    char *ptr;

    if (err) {
	DPRINT(Debug,2,(&Debug,"load_aliases_map: %s: %s (can_open)\n",
			filename,strerror(err)));
	return NULL;
    }

    f = fopen(filename,"r");
    if (!f) {
 	int err UNUSED_VAROK = errno;
	DPRINT(Debug,2,(&Debug,"load_aliases_map: %s: %s\n",
			filename,strerror(err)));
	return NULL;
    }
   
    while(!feof(f) && !ferror(f)) {
        int l1 = malloc_gets(&buf,32000,f);
	char * c = buf;

	if (-1 == l1) {
            lib_error(CATGETS(elm_msg_cat, MeSet, MeTooLongLine,
                              "%.30s: Too long line: %.30s..."),
                      filename,buf ? buf : "???");
            (*errors) ++;
            goto OUT;
	    
        } else if (-2 == l1) {
            int err = errno;
	    
            lib_error(CATGETS(elm_msg_cat, MeSet, MeReadError,
			      "%.30s: Reading error: %s"),
                      filename,strerror(err));
            (*errors) ++;
	    
            goto OUT;
        } else if (0 == l1) {
	    /* There is empty line before different kind alias types */
            continue;
	}
	
	while (l1-- > 0 && whitespace(buf[l1]))
            buf[l1] = '\0';

        while (*c && whitespace ( *c)) /* skip leading whitespace */
            c++;
        if (*c == '#') /* Skip comments */
            continue;
        if (! *c) {
	    /* Empty line exists between different parts of alias defination ...*/

            continue;
	}

	/* Tag line */
	if ('[' == buf[0]) {

	    if (0 == strncmp(ALIAS_TOKEN,buf,strlen(ALIAS_TOKEN))) {
		char * x = buf + strlen(ALIAS_TOKEN);
		struct string * alias_name = NULL;

		while (*x && whitespace ( *x)) /* skip leading whitespace */
		    x++;
		
		if (!*x) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				      "%s: Bad line: %.30s..."),
			      filename,buf);
		    (*errors) ++;
		    continue;
		}

		/* Allow mime encoding used on here .... */
		alias_name = hdr_to_string(HDR_PHRASE,x,cs,1);
		current_node =  give_aliases_map_node(&map_root,alias_name,1);
		free_string(&alias_name);

		alias_count++;

	    } else {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasBadToken,
				  "%s: Bad token on line: %.100s..."),
			  filename,buf);
		
		(*errors) ++;
	    }

	    continue;
	}

	ptr = strchr(buf,':');
	if (!ptr) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasBadLine,
			      "%s: Bad line: %.100s..."),
		      filename,buf);
		
	    (*errors) ++;   
	    continue;
	}
	*ptr = '\0';
	ptr++;
	
	while (whitespace(*ptr))
	    ptr++;

	if (!current_node) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasNameExpected,
			      "%s: Alias name expected: %s: %.100s..."),
		      filename,buf,ptr);
	    (*errors) ++;
	    continue;
	}

	if (0 == strcmp(buf,COMMENT_TOKEN)) {
	    
	    struct string * err = NULL;

	    /* Allow mime encoding used on here .... */
	    struct string * comment = hdr_to_string(HDR_PHRASE,ptr,cs,1);

	    address_alias_add_comment(& current_node->alias, comment, &err);

	    if (err) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError,
				  "%s: %S: %s: %.100s..."),
			  filename,err,buf,ptr);

		(*errors) ++;
		free_string(&err);
	    }
				  
	    free_string(&comment);

	} else if (0 == strcmp(buf,FIRSTNAME_TOKEN)) {
	    
	    struct string * err = NULL;

	    /* Allow mime encoding used on here .... */
	    struct string * firstname = hdr_to_string(HDR_PHRASE,ptr,cs,1);

	    address_alias_add_firstname(& current_node->alias, firstname, &err);

	    if (err) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError,
				  "%s: %S: %s: %.100s..."),
			  filename,err,buf,ptr);

		(*errors) ++;
		free_string(&err);
	    }
				  
	    free_string(&firstname);

	} else if (0 == strcmp(buf,LASTNAME_TOKEN)) {
	    
	    struct string * err = NULL;

	    /* Allow mime encoding used on here .... */
	    struct string * lastname = hdr_to_string(HDR_PHRASE,ptr,cs,1);

	    address_alias_add_lastname(& current_node->alias, lastname, &err);

	    if (err) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError,
				  "%s: %S: %s: %.100s..."),
			  filename,err,buf,ptr);

		(*errors) ++;
		free_string(&err);
	    }
				  
	    free_string(&lastname);


	} else if (0 == strcmp(buf,ADDRESS_TOKEN)) {
	    
	    struct string * err = NULL;

	    /* Allow mime encoding used on here .... */
	    struct address * addr = parse_one_address(ptr,1,cs);

	    if (!addr) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasAddressError,
				  "%s: Failed to parse address: %s: %s"),
			  filename,buf,ptr);
		(*errors) ++;
		continue;
	    }


	    address_alias_add_address(& current_node->alias, addr, &err);

	    if (err) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError,
				  "%s: %S: %s: %.100s..."),
			  filename,err,buf,ptr);

		(*errors) ++;
		free_string(&err);
	    }
				  
	    free_address(&addr);

	} else if (0 == strcmp(buf,LIST_TOKEN)) {
	    
	    struct string * err = NULL;

	    /* Allow mime encoding used on here .... */
	    struct addr_list * list = 
		parse_header_address(NULL,ptr,1,cs,NULL);

	    if (!list) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasAddressListError,
				  "%s: Failed to parse address list: %s: %s"),
			  filename,buf,ptr);
		(*errors) ++;
		continue;
	    }

	    address_alias_add_addr_list(& current_node->alias, list, &err);

	    if (err) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError,
				  "%s: %S: %s: %.100s..."),
			  filename,err,buf,ptr);

		(*errors) ++;
		free_string(&err);
	    }
				  
	    free_addr_list(&list);

	} else if (0 == strcmp(buf,PHRASE_TOKEN)) {
	    
	    struct string * err = NULL;

	    /* Allow mime encoding used on here .... */
	    struct string * phrase = hdr_to_string(HDR_PHRASE,ptr,cs,1);

	    address_alias_add_phrase(& current_node->alias, phrase, &err);

	    if (err) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError,
				  "%s: %S: %s: %.100s..."),
			  filename,err,buf,ptr);

		(*errors) ++;
		free_string(&err);
	    }
				  
	    free_string(&phrase);
	    
	} else if (0 == strcmp(buf,GROUP_TOKEN)) {

	    char *ptr2, *next = NULL;
	    char * temp = safe_strdup(ptr);
	    struct string * err = NULL;

	    address_alias_add_group_start(& current_node->alias,&err);

	    if (err) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError,
				  "%s: %S: %s: %.100s..."),
			  filename,err,buf,ptr);
		
		(*errors) ++;
		free_string(&err);
	    }

	    
	    for (ptr2 = temp; ptr2 && *ptr2; ptr2 = next) {
		struct string * alias = NULL;
		struct string * err = NULL;

		next = qstrpbrk(ptr2,",");
		if (next) {
		    int l2 = next-ptr2;
		    *next = '\0';
		    next++;
		    while (whitespace(*next))
			next++;
		    
		    while (l2-- > 0 && whitespace(ptr2[l2]))
			ptr2[l2] = '\0';
		} 
		
		/* Allow mime encoding used on here .... */
		alias = hdr_to_string(HDR_PHRASE,ptr2,cs,1);

		address_alias_add_group_item(& current_node->alias,
					     alias, &err);
	    
		if (err) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasError2,
				      "%s: %S: %S: %s: %.100s..."),
			      filename,err,alias,buf,ptr);
		    
		    (*errors) ++;
		    free_string(&err);
		}
		
		free_string(&alias);
	    }


	    free(temp);

	} else {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeAliasBadTokenLine,
			      "%s: Bad token line: %s: %.100s..."),
		      filename,buf,ptr);
		
	    (*errors) ++;   
	}

    }
    

    if (ferror(f)) {
        int err = errno;

        lib_error(CATGETS(elm_msg_cat, MeSet, MeReadError,
                          "%s: Reading error: %s"),
                  filename,strerror(err));
        (*errors) ++;
    }

 OUT:
    if (buf) {
        free(buf);
    }
    
    if (!feof(f)) {
        DPRINT(Debug,3,(&Debug,
                        "load_aliases_map: %s, All results not readed.\n",
                        filename));
    }

    fclose(f);

    if (alias_count > 0) {
	DPRINT(Debug,7,(&Debug,"load_aliases_map: %d aliases\n",
			alias_count));

	vector = safe_malloc(alias_count * sizeof (vector[0]));

	fill_vector(map_root,vector,alias_count,&vector_len);


    }

    ret = new_aliases_map();

    ret->root        = map_root;
    ret->vector      = vector;
    ret->vector_len  = vector_len;
    
    return ret;
}

void dump_aliases_map(f,map,commentfile,actor,version_buff)
     FILE *f; 
     struct aliases_map  *map;
     FILE *commentfile;
     const char *actor;
     char *version_buff;
{
    int close_comment_file = 0;
    charset_t  cs  = system_charset;    /* Charset of file */
    int i;

    if (!map)
	return;

    if (ALIASES_map_magic != map->magic)
	panic("URL PANIC",__FILE__,__LINE__,"dump_aliases_map",
	      "Bad magic number",0);
 
    if (!commentfile) {
	commentfile = fopen(ELMALIASES_INFO,"r");   
 	if (!commentfile) {
	    int err = errno;

	    lib_error(CATGETS(elm_msg_cat, MeSet, 
			      MeSavingWithoutComments,
			      "Warning: saving without comments! Can't get to %s: %s"),
		      ELMALIASES_INFO,
		      strerror(err));
	} else 
	    close_comment_file++;
    }
   
    elm_fprintf(f, 
		CATGETS(elm_msg_cat, MeSet, MeSavedAutoWrite1,
			"# Saved automatically by %s %s\n#\n\n"), 
		actor,version_buff);

    if (commentfile) {
	char line_buffer[SLEN];
	int len;

	while (0  < (len = 
		     mail_gets(line_buffer, sizeof line_buffer, 
			       commentfile))) {

	    if (line_buffer[len-1] != '\n') {
		DPRINT(Debug,1,(&Debug, 
				"Too long line (read len=%d): %s\n",
				len,line_buffer));
		break;
	    }
	    line_buffer[--len] = '\0';
	    if (0 == len)
		continue;

	    /* Only copies initial comment */
	    if ('#' == line_buffer[0] || whitespace(line_buffer[0])) {		
		fprintf(f,"%s\n",line_buffer);
	    } else
		break;
	}

	fprintf(f,"\n");
    }

    for (i = 0; i < map->vector_len; i++) {
	struct aliases_map_node * node = map->vector[i];
	char * s1 = NULL;
	
	if (!node)
	    continue;

	if (node->magic != ALIASES_node_magic)
	    panic("ALIASES PANIC",__FILE__,__LINE__,"dump_aliases_map",
		  "Bad node magic number",0);
	
	if (i != node->idx)
	    panic("URL PANIC",__FILE__,__LINE__,"dump_aliases_map",
		  "Bad node index number",0);	

	
	/* Use mime encoding here ...*/
	s1 = string_to_hdr(HDR_PHRASE,node->alias_name,cs,1, NULL,1);

	fprintf(f,"%s\t%s\n",ALIAS_TOKEN,s1); 
	free(s1); s1=NULL;

	if (node->alias) {
	    const struct string * comment = 
		address_alias_get_comment(node->alias);

	    const struct string *firstn   = NULL;
	    const struct string *lastn    = NULL;
	    const struct address *address = NULL;
	    const struct addr_list *list  = 
		address_alias_get_list(node->alias);
	    const struct string * phrase = NULL;
	    const struct alias_vector * group = NULL;
		

	    if (comment) {
		/* Use mime encoding here ...*/
		char * s1 = string_to_hdr(HDR_PHRASE,comment,cs,1, NULL,0);

		fprintf(f,"%s:\t%s\n",
			COMMENT_TOKEN, s1 ? s1 : "");
		free(s1); s1=NULL;
	    }

	    if (address_alias_get_person(node->alias,&firstn,&lastn,
					 &address)) {

		putc('\n',f);

		if (firstn) {
		    /* Use mime encoding here ...*/
		    char * s1 = string_to_hdr(HDR_PHRASE,firstn,cs,1, NULL,1);
		    
		   
		    fprintf(f,"%s:\t%s\n",
			    FIRSTNAME_TOKEN, s1 ? s1 : "");
		    if (s1)
			free(s1); 
		    s1=NULL;
		}

		if (lastn) {
		    /* Use mime encoding here ...*/
		    char * s1 = string_to_hdr(HDR_PHRASE,lastn,cs,1, NULL,1);
		    
		   
		    fprintf(f,"%s:\t%s\n",
			    LASTNAME_TOKEN,s1 ? s1 : "");
		    if (s1)
			free(s1); 
		    s1=NULL;
		}

		if (address) {
		    char * s1 = address_to_str(address,1,cs);
		    
		    fprintf(f,"%s:\t%s\n",
			    ADDRESS_TOKEN,s1 ? s1 : "");
		    if (s1)
			free(s1); 
		    s1=NULL;
		}

	    }

	    if (list) {
		char * s1 = list_to_str(list,1,cs);

		putc('\n',f);
		fprintf(f,"%s:\t%s\n",
			LIST_TOKEN,s1 ? s1 : "");
		if (s1)
		    free(s1); 
		s1=NULL;

	    }

	    if (address_alias_get_group(node->alias,&phrase,&group)) {
		putc('\n',f);

		if (phrase) {
		    /* Use mime encoding here ...*/
		    char * s1 = string_to_hdr(HDR_PHRASE,phrase,cs,1, NULL,1);
		    
		    
		    fprintf(f,"%s:\t%s\n",
			    PHRASE_TOKEN,s1 ? s1 : "");
		    if (s1)
			free(s1); 
		    s1=NULL;
		}
	    
		if (group) {
		    int len = alias_vector_item_count(group);
		    int i;

		    fprintf(f,"%s:\t",GROUP_TOKEN);
		    
		    for (i = 0; i < len; i++) {
			const struct string * alias =
			    alias_vector_get_alias(group,i);

			/* Use mime encoding here ...*/
			char * s1 = string_to_hdr(HDR_PHRASE,alias,cs,1, NULL,1);

			if (i > 0) {
			    putc(',',f);
			}

			if (s1) {
			    fputs(s1,f);
			    free(s1); s1=NULL;
			}
		    }		   

		    putc('\n',f);			
		}
	    }
	}
	
	putc('\n',f);
    }

    if (close_comment_file)
	fclose(commentfile);

}

const struct address_alias * aliases_map_get_alias(map,idx,key)
     const struct aliases_map *map;
     int idx;
     const struct string **key;
{
    if (ALIASES_map_magic != map->magic)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_get_alias",
	      "Bad magic number",0);

    if (idx < 0 || idx >= map->vector_len)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_get_alias",
	      "Bad index",0);

    if (! map->vector[idx])
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_get_alias",
	      "Bad vector",0);

    if (ALIASES_node_magic != map->vector[idx]->magic)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_get_alias",
	      "Bad node magic number",0);

    if (idx != map->vector[idx]->idx)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_get_alias",
	      "Bad node index number",0);

    *key = map->vector[idx]->alias_name;

    return map->vector[idx]->alias;
}

const struct address_alias *  aliases_map_lookup_alias(map,key,idx)
     const struct aliases_map *map;
     const struct string *key;
     int *idx;
{
    const struct aliases_map_node * node = NULL;

    if (ALIASES_map_magic != map->magic)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_lookup_alias",
	      "Bad magic number",0);

    if (idx)
	*idx = -1;

    if (!map->root)
	return NULL;    /* Not found */	

    node = lookup_aliases_map_node(map->root,key);

    if (!node)
	return NULL;    /* Not found */

    if (node->idx >= 0) {
	
	if (node->idx >= map->vector_len)
	    panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_lookup_alias",
		  "bad index",0); 
			       
	if (node != map->vector[node->idx])
	    panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_lookup_alias",
		  "bad vector",0);        

	*idx = node->idx;
    }

    return node->alias;
}


/* Replaces existing aliases */
void aliases_map_add_alias(map,key,alias)
     struct aliases_map *map;
     const struct string *key;
     const struct address_alias *alias;
{
    struct aliases_map_node * node = NULL;
    
    if (ALIASES_map_magic != map->magic)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_add_alias",
	      "Bad magic number",0);

    node = give_aliases_map_node(& (map->root),key,1);

    if (node->idx < 0) {    /* Need add new alias to list */

	map->vector = safe_realloc(map->vector,
				    (map->vector_len+1) *
				    sizeof(map->vector[0]));

	map->vector[map->vector_len] = node;
	node->idx = map->vector_len;

	DPRINT(Debug,9,(&Debug,
			"aliases_map_add_alias: New alias \"%S\" index %d\n",
			key,node->idx));

	map->vector_len++;

    } else {

	if (node->idx >= map->vector_len)
	    panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_add_alias",
		  "bad index",0); 
	
	if (node != map->vector[node->idx])
	    panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_add_alias",
		  "bad vector",0);        

	DPRINT(Debug,15,(&Debug,
			"aliases_map_add_alias: Updating alias \"%S\" index %d\n",
			key,node->idx));

    }

    if (node->alias)
	free_address_alias( & (node->alias));

    node->alias = dup_address_alias(alias);

    return;
}

int aliases_map_remove_alias(map,key) 
     struct aliases_map *map;
     const struct string *key;
{
    struct aliases_map_node * node = NULL;

    if (ALIASES_map_magic != map->magic)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_remove_alias",
	      "Bad magic number",0);

    node = give_aliases_map_node(& (map->root),key, 0);

    if (!node)
	return 0;    /* Not found */

    if (node->alias) {
	if (node->idx < 0)
	    panic("URL PANIC",__FILE__,__LINE__,"aliases_map_remove_alias",
		  "No index for alias",0); 

	free_address_alias( & (node->alias));
    }

    if (node->idx >= 0) {
	int i;

	if (node->idx >= map->vector_len)
	    panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_remove_alias",
		  "bad index",0); 
	
	if (node != map->vector[node->idx])
	    panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_remove_alias",
		  "bad vector",0);      
	
	for (i = node->idx+1; i < map->vector_len; i++) {

	    if (ALIASES_node_magic != map->vector[i]->magic)
		panic("URL PANIC",__FILE__,__LINE__,"aliases_map_remove_alias",
		      "Bad node magic number",0);
	    
	    if (i != map->vector[i]->idx)
		panic("URL PANIC",__FILE__,__LINE__,"aliases_map_remove_alias",
		      "Bad node index number",0);	    
	    
	    map->vector[i-1] = map->vector[i];
	    map->vector[i-1]->idx = i-1;
	    map->vector[i] = NULL;
	}
	map->vector_len--;
	node->idx = -1;                /* Node not in use */

	return 1;
    }

    return 0;
}

void aliases_map_remove_index(map,idx)  
      struct aliases_map *map;
      int idx;
{
    int i;
    struct aliases_map_node * node = NULL;

    if (ALIASES_map_magic != map->magic)
	panic("URL PANIC",__FILE__,__LINE__,"aliases_map_remove_index",
	      "Bad magic number",0);
    
    if (idx < 0 || idx >= map->vector_len)
	panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_remove_index",
	      "bad index",0); 
    
    node = map->vector[idx];
    
    if (idx != node->idx)
	if (node != map->vector[node->idx])
	    panic("ALIASES PANIC",__FILE__,__LINE__,"aliases_map_remove_index",
		  "bad vector",0);    
    
    for (i = idx+1; i < map->vector_len; i++) { 
	
	if (ALIASES_node_magic != map->vector[i]->magic)
	    panic("URL PANIC",__FILE__,__LINE__,"aliases_map_remove_alias",
		  "Bad node magic number",0);
	
	if (i != map->vector[i]->idx)
	    panic("URL PANIC",__FILE__,__LINE__,"aliases_map_remove_index",
		  "Bad node index number",0);	    
	
	map->vector[i-1] = map->vector[i];
	map->vector[i-1]->idx = i-1;
	map->vector[i] = NULL;
    }
    map->vector_len--;
    node->idx = -1;                /* Node not in use */

    free_address_alias( & (node->alias));
}


int aliases_map_item_count(map)
     struct aliases_map *map;
{
    return map->vector_len;
} 

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
