static char rcsid[] = "@(#)$Id: smtp.c,v 2.3 2014/11/02 12:20:45 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elmtls.h"
#include "shared_smtp.h"

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

#ifdef REMOTE_MBX

DEBUG_VAR(Debug,__FILE__,"tls");

S_(smtp_ext_handler smtp_STARTTLS_handler)
static  int smtp_STARTTLS_handler P_((struct mailer_info_1 *mailer_info,
				      char * ext,
				      char * ext_args,
				      size_t mailer_info_size,
				      struct smtp_callbacks *cb, 
				      size_t cb_size,
				      struct SE_option   *service_config,
				      size_t sc_size,
				      enum CAPA_phase *phase));
static  int smtp_STARTTLS_handler(mailer_info,ext,ext_args,mailer_info_size,
				  cb,cb_size,service_config,sc_size,phase)
     struct mailer_info_1  * mailer_info;
     char                  * ext;
     char                  * ext_args;
     size_t                  mailer_info_size;
     struct smtp_callbacks * cb;
     size_t                  cb_size;
     struct SE_option      * service_config;
     size_t                  sc_size;
     enum CAPA_phase       * phase;
{
    union stream_types T;   

    int ret = 1;
    int tls = 1;
    int bits;
    char response_code[4];
    char * M = NULL;
    
    if (! StreamOK(sizeof (struct stream_type))) {
	DPRINT(Debug,1,(&Debug,			    
			"tls:smtp_STARTTLS_handler: sizeof (struct stream_type) mismatch\n"));
	
	/* do nothing  
	   -- not need to return failure 
	   -- just disable ourself
	*/
	goto fail;
    }


    if (mailer_info_size != sizeof (*mailer_info)) {
	DPRINT(Debug,1,(&Debug,			    
			"tls:smtp_STARTTLS_handler: sizeof (struct mailer_info_1) = %d != %d\n",
			sizeof (*mailer_info),mailer_info_size));
	
	/* do nothing  
	   -- not need to return failure 
	   -- just disable ourself
	*/
	goto fail;
    }

    if (cb_size != sizeof (*cb)) {
	DPRINT(Debug,1,(&Debug,			    
			"tls:smtp_STARTTLS_handler: sizeof (struct smtp_callbacks) = %d != %d\n",
			sizeof (*cb),cb_size));
	
	/* do nothing  
	   -- not need to return failure 
	   -- just disable ourself
	*/
	goto fail;
    }

    if (sc_size != sizeof (*service_config)) {
	DPRINT(Debug,1,(&Debug,			    
			"tls:smtp_STARTTLS_handler: sizeof (struct struct SE_option) = %d != %d\n",
			sizeof (*service_config),sc_size));
	
	/* do nothing  
	   -- not need to return failure 
	   -- just disable ourself
	*/
	goto fail;
    }
    
    if (!remote_account_OK (sizeof (mailer_info->RA))) {
	DPRINT(Debug,1,(&Debug,			    
			"tls:smtp_STARTTLS_handler: sizeof (struct remote_account) mismatch\n"));
	
	/* do nothing  
	   -- not need to return failure 
	   -- just disable ourself
	*/
	goto fail;
    }

    if (!tls_init(NULL)) {
	DPRINT(Debug,1,(&Debug,			    
			"tls:smtp_STARTTLS_handler: Initialization failed\n"));
	goto fail;
    }

    StreamInfo(mailer_info->RA.stream,SS_ssf,&bits,NULL,NULL);
    if (bits > 0) {
	DPRINT(Debug,4,(&Debug,			    
			"tls:smtp_STARTTLS_handler: Already %d 'bits' encryption -- TLS already active?\n",
			bits));
	goto fail;
    }   

    if (service_config) {
	if (service_config->type != &tls_options ||
	    !service_config->value) {
	    panic("TLS PANIC",__FILE__,__LINE__,"smtp_STARTTLS_handler",
		  "Bad option storage",0);
	}
	
	if (tls_none == service_config->value->v_starttls) {
	    DPRINT(Debug,4,(&Debug,			    
			    "tls:smtp_STARTTLS_handler: tls disabled\n"));
	    goto fail;
	}
	    

	tls = service_config->value->v_starttls;
    }

    if (!cb->start_command(mailer_info,"STARTTLS")) {
	cb->command_reset(mailer_info);
	goto fail;            /* Just fail ... */
    }
    cb->command_ready(mailer_info,0);  /* Command do not have arguments */

    if (!cb->command_ok(mailer_info,response_code)) {
	if (response_code[0]) {
	    int idx = 0;
	    char * r = 
		cb->command_response(mailer_info,&idx);
	    
	    if (r) {
		lib_error(FRM("%s server rejected STARTTLS with %3.3s %s"),
			  (mailer_info->SMTP_flags & 
			   SMTP_submission) ?
			  "Submission" : "SMTP",
			  response_code,r);

		while (NULL != (r = 
				cb->command_response(mailer_info,
						     &idx))) {
		    lib_error(FRM("%s server: %s"),
			      (mailer_info->SMTP_flags & 
			       SMTP_submission) ?
			      "Submission" : "SMTP",
			      r);
		}
	    }
	}
	
	cb->command_reset(mailer_info);
	goto fail;            /* Just fail ... */
    }

    if (0 != memcmp(response_code,"220",3)) {
	DPRINT(Debug,5,(&Debug,
			 "tls:smtp_STARTTLS_handler: not expected response code: %.3s\n",
			response_code));

	ret = 0;  /* Hard failure -- not sure if connection is usefull */
	goto fail;
    }

    cb->command_reset(mailer_info);
    
    /* After that we need need return failure if something
       goes wrong because connection is messed beyond repair
    */

    T = create_TLS_stream(tls);
    if (T.TYPE == NULL) {
	/* Error message ? */
	ret = 0;
	goto fail;
    }

    lib_transient(FRM("%s/%s connect to %s ..."),
		  (mailer_info->SMTP_flags & SMTP_submission) ?
		  "Submission" : "SMTP",
		  tls ? "TLS" : "SSL",
		  mailer_info->RA.host);

    add_stream_to_head(mailer_info->RA.stream,T);

    /* We are head on stream */  

    /* Do handshake */
    while (tls_not_connected == 
	   ss_TLS_state(mailer_info->RA.stream,0)) {
	WaitStreamFor(mailer_info->RA.stream,
		      SS_setup_act);
	M =  RemoveStreamError(mailer_info->RA.stream);
	if (M)
	    goto error_jump;
    }

    switch(ss_TLS_state(mailer_info->RA.stream,0)) {
	struct string * errmsg;
	struct string * peer_name;
    
    default: error_jump:
    
	if (M) {
	    /* Error message ? */
	    lib_error(FRM("%s"),
		      M);
	    free(M);
	    M = NULL;
	} else {
	    /* Error message ? */
	    lib_transient(FRM("%s/%s connect to %s ... Fail"),
			  (mailer_info->SMTP_flags & SMTP_submission) ?
			  "Submission" : "SMTP",
			  tls ? "TLS" : "SSL",
			  mailer_info->RA.host);
	}
	 
	/* We fail hard now if connection is not OK ... */
	ret = 0;
	goto fail;

    case tls_connected:

	/* Message ????? */

	errmsg     = NULL;
	peer_name = NULL;

	switch(ss_TLS_verify(mailer_info->RA.stream,0,
			     &errmsg,&peer_name)) {

	default:
	    if (errmsg)
		lib_transient(FRM("%s/%s connect to %s ... Done (%S)"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host,
			      errmsg);
	    else
		lib_transient(FRM("%s/%s connect to %s ... Done"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host);
	    break;
	    
	case tls_not_verified:
	    if (errmsg)
		lib_transient(FRM("%s/%s connect to %s ... Done (not verified; %S)"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host,
			      errmsg);
	    else
		lib_transient(FRM("%s/%s connect to %s ... Done (not verified)"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host);
	    break;

	case tls_peer_verified:
	    if (peer_name)
		lib_transient(FRM("%s/%s connect to %s ... Done (verified as %S)"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host,
			      peer_name);
	    else
		lib_transient(FRM("%s/%s connect to %s ... Done (verified)"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host);
	    break;

	case tls_verify_failed:
	    if (errmsg)
		lib_transient(FRM("%s/%s connect to %s ... Done (verify failed: %S)"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host,
			      errmsg);
	    else
		lib_transient(FRM("%s/%s connect to %s ... Done (verify failed)"),
			      (mailer_info->SMTP_flags & SMTP_submission) ?
			      "Submission" : "SMTP",
			      tls ? "TLS" : "SSL",
			      mailer_info->RA.host);
	    break; 
	}

	if (errmsg)
	    free_string(&errmsg);
	if (peer_name)
	    free_string(&peer_name);
    
	if (sleepmsg)
	    wait_for_timeout(sleepmsg);

	/* Need retry EHLO command */
	if (capa_prelogin == *phase)
	    *phase = capa_prelogin_again;
    }
    
 fail:
    return ret;
}

static struct provides_shared_SMTP_EXT LIST[] = {
  { "STARTTLS", 1, smtp_STARTTLS_handler },
};

#ifdef ANSI_C
provides_shared_SMTP_EXT_f provides_shared_SMTP_EXT;
#endif
struct provides_shared_SMTP_EXT  *  provides_shared_SMTP_EXT(count,s_size)
     int *count; 
     size_t *s_size;
{
    *s_size = sizeof (LIST[0]);

    *count = (sizeof LIST) / *s_size;

    return &(LIST[0]);
}


#endif

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
